﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace FractalCompression
{
    class Chromosome:IComparable<Chromosome>    
    {
        //mozda bolje da se gen alg vrti posebno za svaki range block
        public FractalMap[] _mappings;
        static double _errorTreshold=0;
        public static bool UseSecondFitness = false;
        FractalMap _maxValue;
        int _shiftWindow=16;
        double _fitness;
        static double _mutationP=0.15;
        static double _crossoverP=0.8;
        int _numBlocksUnderTreshold = 0;
        Random _R=new Random();
        public Chromosome(FractalMap maxValue,int noMappings)
        {
            _maxValue=maxValue;
            _mappings = new FractalMap[noMappings];
           // Random R=new Random();
            for (int i = 0; i < noMappings; i++)
            {
                FractalMap temp = new FractalMap();
                temp.LumShift = _R.Next(_maxValue.LumShift);
                temp.X = _R.Next(_maxValue.X);
                temp.Y = _R.Next(_maxValue.Y);
                temp.Isometry = _R.Next(_maxValue.Isometry);
                temp.Contrast = 0.5;//R.NextDouble() * 1.5;
                temp.Error = double.MaxValue;
                _mappings[i]=temp;
            }
            _fitness = Util.EvaluateMappings2(_mappings);
        }
        public Chromosome(Chromosome parent1, Chromosome parent2)
        {
            _mappings = new FractalMap[parent1._mappings.Length];
            _maxValue = parent1._maxValue;
            if (_R.NextDouble() < _crossoverP)
               //  onePointCrossover(parent1, parent2);
                smartCrossover(parent1, parent2);
            else
            {
                for (int i = 0; i < _mappings.Length; i++)
                    this._mappings[i] = new FractalMap(parent1._mappings[i]);
            }
            mutate();
            _fitness=Util.EvaluateMappings2(_mappings);

        }

        private void onePointCrossover(Chromosome parent1, Chromosome parent2)
        {
            int x = parent1._mappings.Length/ 2;
            for (int i = 0; i < x; i++)
            {
                _mappings[i] = new FractalMap(parent1._mappings[i]);
            }
            for (int i = x; i < parent1._mappings.Length; i++)
            {
                _mappings[i] = new FractalMap(parent2._mappings[i]);
            }

            //for (int i = 0; i < x; i++)
            //{
            //    _mappings[i] = parent1._mappings[i];
            //}
            //for (int i = x; i < parent1._mappings.Length; i++)
            //{
            //    _mappings[i] = parent2._mappings[i];
            //}
        }

        private void smartCrossover(Chromosome parent1, Chromosome parent2)
        {
            for (int i = 0; i < parent1._mappings.Length; i++)
            {
                if (parent1._mappings[i].Error < parent2._mappings[i].Error)
                    _mappings[i] = new FractalMap(parent1._mappings[i]);
                else
                    _mappings[i] = new FractalMap(parent2._mappings[i]);

            }
        }


        private int mutateValue(int value, int max, Random R)
        {
            int shift = _shiftWindow / 2 - R.Next(_shiftWindow);
            value += shift;
            if (value >= max|| value < 0)
            {
                value = Math.Abs(value - max) % max;
            }
            return value;

        }

        private void mutate2()
        {
            for (int i = 0; i < _mappings.Length; i++)
            {
                if (_mappings[i].Error < _errorTreshold)
                {
                    _numBlocksUnderTreshold++;
                    continue;
                }
                if (_R.NextDouble() < _mutationP)
                {
                    _mappings[i].X = _R.Next(_maxValue.X);
                    _mappings[i].Y = _R.Next(_maxValue.Y);
                    _mappings[i].Isometry = _R.Next(_maxValue.Isometry);
                }
            }
        }


        //sto s kontrastom?
        private void mutate()
        {           
            _numBlocksUnderTreshold = 0;
            for (int i = 0; i < _mappings.Length; i++)
            {
                //FractalMap map = _mappings[i];
                if (_mappings[i].Error < _errorTreshold)
                {
                    _numBlocksUnderTreshold++;
                    continue;
                }
                if (_R.NextDouble() < _mutationP)
                {
                   switch(_R.Next(3))
                   {
                       case 0: _mappings[i].X = mutateValue(_mappings[i].X, _maxValue.X, _R); break;
                       case 1: _mappings[i].Y = mutateValue(_mappings[i].Y, _maxValue.Y, _R); break;
                       case 2: _mappings[i].Isometry = mutateValue(_mappings[i].Isometry, _maxValue.Isometry, _R); break;
                    //   case 4: _mappings[i].LumShift = mutateValue(_mappings[i].LumShift, _maxValue.LumShift, R); break;
                   }
                }
            }

        }

        public int CompareTo(Chromosome C)
        {
            if (this.Fitness - C.Fitness < 0) return 1;
            if (this.Fitness == C.Fitness) return 0;
            return -1;
        }



        
        //za kompaktniji zapis kad ce se zapisivati u datoteku
        public void Compact(byte[] bytes,ref int bytePos,ref int bitPos)
        {
            for (int i = 0; i < _mappings.Length; i++)
            {
                _mappings[i].Compact(bytes, ref bytePos, ref bitPos);
            }
        }
        public double Fitness
        {
            get
            {
                if (!UseSecondFitness)
                    return _fitness;
                else return _numBlocksUnderTreshold;
            }
        }
        public int Fitness2
        {
            get
            {
                return _numBlocksUnderTreshold;
            }
        }
        public FractalMap[] Mappings
        {
            get
            {
                return _mappings;
            }
        }
        public static double MutationProbability
        {
            get
            {
                return _mutationP;
            }
            set
            {
                _mutationP = value;
            }
        }
        public static double CrossoverProbability
        {
            get
            {
                return _crossoverP;
            }
            set
            {
                _crossoverP = value;
            }
        }
        public static double ErrorThreshold
        {
            get
            {
                return _errorTreshold;
            }
            set
            {
                _errorTreshold = value;
            }
        }





    }
}

