﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;

namespace FractalCompression
{
    
    public class FractalEncoder
    {
        Bitmap img;
        Size _rangeSize = new Size(4, 4);
        Size _domainSize = new Size(8, 8);
        int _generationNumber = 30;
        int _stagnationThreshold=0;
        int _populationSize = 50;
        string _outputPath;
        public event EventHandler ProgressChanged;
        double _progress;
        int numRanges;
        public int Progress
        {
            get
            {
                return (int)(_progress*100)>100?100:(int)(_progress*100);
            }
        }

        public FractalEncoder(string filePath,string outputPath)
        {
            img= new Bitmap(filePath);
            _outputPath=outputPath;
        }
        public void Encode()
        {
            BinaryWriter bw = new BinaryWriter(new FileStream(_outputPath, FileMode.Create));

            Chromosome best = null;
            int ind = 0;
            numRanges = (img.Width / Util._rangeSize.Width) * (img.Height / Util._rangeSize.Height);
            int numRangesInPictureBlock = (Util.PictureBlockSize.Width * Util.PictureBlockSize.Height) / (_rangeSize.Width * _rangeSize.Height);
            bw.Write((short)(img.Width / Util._rangeSize.Width));
            bw.Write((short)(img.Height / Util._rangeSize.Height));
            bw.Write((short)(Util.PictureBlockSize.Width / Util._rangeSize.Width));


            byte[] compacted = new byte[(numRanges * FractalMap.CompactLength) / 8 + 1];
            int PictureBlockXcoord = 0;
            int PictureBlockYcoord = 0;

            int bytePos = 0, bitPos = 0;
            while (PictureBlockYcoord < img.Height)
            {
                Util.GetPixelsFromBmp(img, PictureBlockXcoord, PictureBlockYcoord);
                PictureBlockXcoord += Util.PictureBlockSize.Width;
                if (PictureBlockXcoord >= img.Width)
                {
                    PictureBlockXcoord = 0;
                    PictureBlockYcoord += Util.PictureBlockSize.Height;

                }
                double weight = (double)Util.numberOfRangeBlock / numRangesInPictureBlock;
                best = getBestForBlock(Util.numberOfRangeBlock);
                ind++;
                best.Compact(compacted, ref bytePos, ref bitPos);
            }
            bw.Write(compacted);
            bw.Close();
        }



        private Chromosome getBestForBlock(int noRanges)
        {
            Population pop = new Population(_populationSize,noRanges);
            int numGenerations = _generationNumber;
            int iterationsNotChanged = 0;
            double currentFitness = 0;

            for (int i = 0; i < numGenerations; i++)
            {
                pop.AdvanceGeneration();


                double delta = ((double)1 / numGenerations) * noRanges / numRanges;
                double progressOld = _progress;
                _progress = _progress + delta;
                if (Math.Floor(progressOld*100) < Math.Floor(_progress*100))
                    ProgressChanged(this, null);


                Console.WriteLine("{0}. generacija Prosjek: {1}", i + 1, pop.AverageFitness);
                Console.WriteLine("Najbolja: {0}", pop.Individuals[0].Fitness);
                Console.WriteLine("Najbolja drugi fitnes: {0}", pop.Individuals[0].Fitness2);
                if (pop.Individuals[0].Fitness != currentFitness)
                {
                    iterationsNotChanged = 1;
                    currentFitness = pop.Individuals[0].Fitness;
                }
                else iterationsNotChanged++;
                if (_stagnationThreshold!=0 && iterationsNotChanged == StagnationThreshold)
                {
                    break;
                    //iterationsNotChanged = 1;
                    //   Chromosome.UseSecondFitness = true;
                    // Chromosome._errorTreshold -= 2;
                    // break;

                }


            }
            Console.WriteLine("max contrast {0}", Util.maxContrast);
            Console.WriteLine("max bright {0}", Util.maxBright);
            return pop.Individuals[0];
        }

        public Size RangeBlockSize
        {
            get
            {
                return _rangeSize;
            }
            set
            {
                _rangeSize = value;
                Util._rangeSize = _rangeSize;
            }
        }
        public Size DomainBlockSize
        {
            get
            {
                return _domainSize;
            }
            set
            {
                _domainSize = value;

                Util._domainSize = _domainSize;
            }
        }
        public double MutationProbability
        {
            get
            {
                return Chromosome.MutationProbability;
            }
            set
            {
                Chromosome.MutationProbability = value;
            }
        }
        public double CrossoverProbability
        {
            get
            {
                return Chromosome.CrossoverProbability;
            }
            set
            {
                Chromosome.CrossoverProbability = value;
            }
        }
        public double ErrorThreshold
        {
            get
            {
                return Chromosome.ErrorThreshold;
            }
            set
            {
                Chromosome.ErrorThreshold = value;
            }
        }
        public int GenerationNumber
        {
            get
            {
                return _generationNumber;
            }
            set
            {
                _generationNumber = value;
            }
        }
        public int StagnationThreshold
        {
            get
            {
                return _stagnationThreshold;
            }
            set
            {
                _stagnationThreshold = value;
            }
        }
        public int PopulationSize
        {
            get
            {
                return _populationSize;
            }
            set
            {
                _populationSize = value;
            }
        }

                


    }
}
