﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace FractalCompression
{
    public class FractalMap:IComparable<FractalMap>
    {
        public int X;
        public int Y;
        public int Isometry;
        public int LumShift;
        public double Contrast;
        public double Error;
        public const int CompactLength = 36;
        public FractalMap()
        {
            Contrast = 0.5;
        }

        private ushort readValue(byte[] bytes,ref int bytePos,ref int bitPos,int bitCount)
        {
            uint temp;
            byte[] tempB = new byte[4];
            if(bytePos + 3<bytes.Length)tempB[0] = bytes[bytePos + 3];
            if (bytePos + 2 < bytes.Length) tempB[1] = bytes[bytePos + 2];
            if (bytePos + 1 < bytes.Length) tempB[2] = bytes[bytePos + 1];
            tempB[3] = bytes[bytePos];
            temp=BitConverter.ToUInt32(tempB,0);
            temp=temp<<bitPos;
            temp=temp>>(32-bitCount);

            bitPos+=bitCount;
            if(bitPos>=8)
            {
                bytePos += bitPos / 8;
                bitPos=bitPos%8;            
            }

            return (ushort)temp;
        }

        public FractalMap(byte[] bytes,ref int bytePos,ref int bitPos)
        {
            X = readValue(bytes,ref  bytePos,ref bitPos, 7);
            Y = readValue(bytes, ref  bytePos, ref bitPos, 7);
            Isometry = readValue(bytes, ref  bytePos, ref bitPos, 3);
            LumShift = readValue(bytes, ref  bytePos, ref bitPos, 9) -256;
            Contrast = readValue(bytes, ref  bytePos, ref bitPos, 10);// -1024;
            Contrast = Contrast / (1 << 10);
        }

        public FractalMap(FractalMap map)
        {
            X = map.X;
            Y = map.Y;
            Isometry = map.Isometry;
            LumShift = map.LumShift;
            Contrast = map.Contrast;
            Error = map.Error;
        }

        public override string ToString()
        {
            return "X:" + X.ToString() +
                " Y:" + Y.ToString() +
                " Iso:" + Isometry.ToString() + " Lum:" + LumShift.ToString() + " Con:" + Contrast.ToString();
        }

        private void writeToBuffer(byte[] buffer,ref int bytePos,ref int bitPos,byte[] bytes, int bitCount,int byteCount)
        {
            for(int i=0;i<byteCount;i++)
            {
                byte leftPart, rightPart;
                
                leftPart = (byte)(bytes[i] >> bitPos);
                rightPart = (byte)(bytes[i] << (8 - bitPos));

              
                buffer[bytePos] |= leftPart;
                if(bytePos+1<buffer.Length)buffer[bytePos+1] |= rightPart;
                int count = bitCount > 8 ? 8 : bitCount;
                if (bitPos + count >= 8)
                    bytePos++;

                bitPos = (bitPos + count) % 8;
                bitCount -= 8;
                if (bitCount <= 0) break;
            }
        }

        public byte[] Compact()
        {
            return null;
        }

        public void Compact(byte[] buffer,ref int bytePos,ref int bitPos)
        {
            List<byte> bytes = new List<byte>();

            byte[] temp = new byte[2];
            temp[0]=(byte)(X<<1);
            writeToBuffer(buffer,ref bytePos,ref bitPos,temp,7,1);
            temp[0]=(byte)(Y<<1);
            writeToBuffer(buffer,ref bytePos,ref bitPos,temp,7,1);
            temp[0] = (byte)(Isometry << 5);
            writeToBuffer(buffer, ref bytePos, ref bitPos, temp, 3,1);
            uint shift = (uint)((LumShift + 256)<<7);
            temp[0] = (byte)(shift >> 8);
            temp[1] = (byte)(shift);
            writeToBuffer(buffer, ref bytePos, ref bitPos, temp, 9, 2);
            int finContrast = (int)(Contrast * (1 << 10));
            if (Math.Abs(finContrast) >= 1023) finContrast = Math.Sign(finContrast) * 1023;
            uint con = (uint)((finContrast)<<6);
            temp[0] = (byte)(con >> 8);
            temp[1] = (byte)(con);
            writeToBuffer(buffer, ref bytePos, ref bitPos, temp, 10, 2);


        }

        public int CompareTo(FractalMap m)
        {
            if (m.X != X || m.Y != Y || m.LumShift != LumShift || m.Isometry != Isometry || m.Contrast != Contrast) return -1;
            return 0;
        }
        


    }
}
