﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace FractalCompression
{
    class Population
    {
        List<Chromosome> _theBest = new List<Chromosome>();
        List<Chromosome> _individuals = new List<Chromosome>();
        int _size;
        int _tournamentSize=5;
        int _elitismChromoCount=3;
        double _averageFitness;
        bool _useSecondFitness;
        Random _R=new Random();
        int noMappings;
        public Population(int size,int numMappings)
        {
            _size = size;
            FractalMap max=new FractalMap();
            max.X=Util.PictureBlockSize.Width-Util._domainSize.Width+1;
            max.Y=Util.PictureBlockSize.Height-Util._domainSize.Height+1;
            max.Isometry=8;
            max.LumShift=128;
            max.Contrast=2;
            for (int i = 0; i < size; i++)
            {
                _individuals.Add(new Chromosome(max, numMappings));
            }
            _individuals.Sort();
            
            _theBest.AddRange(_individuals.Take(_elitismChromoCount));
            
            
        }

        private IEnumerable<Chromosome> getPotentialParents(Random R)
        {
            List<Chromosome> potentialParents = new List<Chromosome>();
            for(int i=0;i<_tournamentSize;i++)
            {
                potentialParents.Add(_individuals[R.Next(_size)]);
            }
            return potentialParents;
        }

        public void AdvanceGeneration()
        {
            List<Chromosome> newIndividuals = new List<Chromosome>();
            List<Chromosome> potentialParents=new List<Chromosome>();
            newIndividuals.AddRange(_theBest);
            _theBest.Clear();
            _averageFitness = 0;
            for(int i=0;i<_individuals.Count-_elitismChromoCount;i++)
            {
                potentialParents.Clear();
                potentialParents.AddRange(getPotentialParents(_R));
                potentialParents.Sort();
                //Chromosome parent1=potentialParents[_R.Next(_tournamentSize)];
                //potentialParents.Remove(parent1);
                //Chromosome parent2=potentialParents[_R.Next(_tournamentSize-1)];

                Chromosome parent1 = potentialParents[0];
                potentialParents.Remove(parent1);
                Chromosome parent2 = potentialParents[0];
                Chromosome child = new Chromosome(parent1, parent2);
                newIndividuals.Add(child);
                
                
            }
            for (int i = 0; i < _size; i++)
            {
                _averageFitness += _individuals[i].Fitness;
            }
            _averageFitness /= _size;
            newIndividuals.Sort();
            _theBest.AddRange(newIndividuals.Take(_elitismChromoCount));
            _individuals=newIndividuals;
                

        }
        public double AverageFitness
        {
            get
            {
                return _averageFitness;
            }
        }
        public List<Chromosome> Individuals
        {
            get
            {
                return _individuals;
            }
        }
        public int NumberOfMappings
        {
            get
            {
                return noMappings;
            }
            set
            {
                noMappings = value;
            }
        }
        public bool UseSecondFitnessForEvaluation
        {
            get
            {
                return _useSecondFitness;
            }
            set
            {
                _useSecondFitness = value;
            }
        }

    }
}
