﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Drawing;
using System.IO;
using System.Drawing.Imaging;
using System.Text;

namespace FractalCompression
{
    public class Util
    {

        static byte[,] outPixels = new byte[128, 128];
        static byte[,] temp = new byte[128, 128];
        public static Size PictureBlockSize=new Size(128,128);
        public static Size _rangeSize = new Size(4, 4);
        public static Size _domainSize=new Size(8,8);
        static Bitmap img;// = new Bitmap("Untitled2.bmp");
        static byte[,] pixels;
        static int noIterations = 10;
        public static double maxContrast;
        public static double maxBright;
        public static int numberOfRangeBlock;
        static Size realPictureBlockSize;

    /*    public static int NumberOfRangeBlocks
        {
            get
            {
                return numberOfRangeBlock;
            }
        }*/
        public static double ComputeAveragePixelError(string originalFilePath, string outputFilePath)
        {
            Bitmap input = new Bitmap(originalFilePath);
            Bitmap output = new Bitmap(outputFilePath);
        

            double error = 0;
            if (input.Size != output.Size)
            {
                throw new Exception("Size does not match");
            }
            for (int i = 0; i < input.Width; i++)
            {
                for (int j = 0; j < input.Height; j++)
                {
                    error += Math.Abs(input.GetPixel(i, j).R - output.GetPixel(i, j).R);
                }
            }
            
            error = error / (input.Height * input.Width);
            return error;
        }
        public static double ComputePeakSignaltoNoise(string originalFilePath, string outputFilePath)
        {
            Bitmap input = new Bitmap(originalFilePath);
            Bitmap output = new Bitmap(outputFilePath);


            double error = 0;
            if (input.Size != output.Size)
            {
                throw new Exception("Size does not match");
            }
            for (int i = 0; i < input.Width; i++)
            {
                for (int j = 0; j < input.Height; j++)
                {
                    error += Math.Pow(Math.Abs(input.GetPixel(i, j).R - output.GetPixel(i, j).R),2);
                }
            }

            error = error / (input.Height * input.Width);


            double rms=Math.Sqrt(error);
            
            double PSNR=20*Math.Log10(255/rms);
            return PSNR;
        }

        
        public static double ComputeAveragePixelError()
        {
           

            double error = 0;
            
            for (int i = 0; i < img.Height; i++)
            {
                for (int j = 0; j < img.Width; j++)
                {
                    error+= Math.Abs(pixels[i, j] - outPixels[i, j]);
                }
            }

            error = error / (img.Width*img.Height);
            return error;
        }
        public static double ComputePeakSignaltoNoise()
        {
            double error = 0;

            for (int i = 0; i < img.Height; i++)
            {
                for (int j = 0; j < img.Width; j++)
                {
                    error +=Math.Pow( Math.Abs(pixels[i, j] - outPixels[i, j]),2);
                }
            }

            error = error / (img.Width * img.Height);

            double rms = Math.Sqrt(error);

            double PSNR = 20 * Math.Log10(255 / rms);
            return PSNR;
        }


        public static double EvaluateMappings2(FractalMap[] mappings)
        {
            double err = 0;
            for (int i = 0; i < mappings.Length; i++)
            {
                //  byte[,] domainBlock = new byte[_domainSize.Height, _domainSize.Width];
                //getDomainBlock(outPixels,Mappings[i].Y,Mappings[i].X ,domainBlock);

                byte[,] rangeBuffer = transformToRangeBlock(pixels, mappings[i].X, mappings[i].Y, mappings[i].Isometry/* Mappings[i].LumShift*/);

                double contrast = 0;
                int bright = 0;
                byte[,] rangeBlock = getRangeBlock(pixels, i);
                getParameters(rangeBlock, rangeBuffer, out contrast, out bright);
                if (Math.Abs(contrast) > maxContrast)
                {
                    maxContrast = contrast;
                }
                if (Math.Abs(bright) > maxBright)
                {
                    maxBright = bright;
                }
                mappings[i].LumShift = bright;
                mappings[i].Contrast = contrast;
                mappings[i].Error = MSEError(rangeBlock, rangeBuffer, contrast, bright);
              /*  if (mappings[i].Error > err)
                {
                    err = mappings[i].Error;
                }*/
                err += mappings[i].Error;
            }
            return 1/(1+err / mappings.Length);
           // return 1 / (1 + err );
          //  return ComputePeakSignaltoNoise();
        }



        public static void GetPixelsFromBmp(Bitmap bmp,int x, int y)
        {

            
            pixels = new byte[PictureBlockSize.Height, PictureBlockSize.Width];
            //x *= PictureBlockSize.Width;
         //   y *= PictureBlockSize.Height;
            int startX = x + PictureBlockSize.Width > bmp.Width ? bmp.Width - PictureBlockSize.Width : x;
            int startY = y + PictureBlockSize.Height > bmp.Height ? bmp.Height- PictureBlockSize.Height : y;
            realPictureBlockSize = new Size((PictureBlockSize.Width - startX % PictureBlockSize.Width)
                , ( PictureBlockSize.Height - startY % PictureBlockSize.Height));
            numberOfRangeBlock = (realPictureBlockSize.Width * realPictureBlockSize.Height) / (_rangeSize.Width * _rangeSize.Height);
            for (int i = 0; i < PictureBlockSize.Height; i++)
            {
                for (int j = 0; j < PictureBlockSize.Width; j++)
                {
                   // Color clr = bmp.GetPixel(j+startX, i);
                    pixels[i, j] = bmp.GetPixel(j+startX, i+startY).R;
                }
            }
            //return pixels;

        }
        private static Point getRangeBlockCoordinates(int i)
        {
            int x = (i * _rangeSize.Width) % (realPictureBlockSize.Width)+(PictureBlockSize.Width-realPictureBlockSize.Width);
            int y = ((i * _rangeSize.Width) / realPictureBlockSize.Width) * _rangeSize.Height + (PictureBlockSize.Height - realPictureBlockSize.Height);
            return new Point(x, y);

        }
        private static byte[,] getRangeBlock(byte[,] input, int index)
        {
            Point coords = getRangeBlockCoordinates(index);
            byte[,] rangeBlock = new byte[_rangeSize.Height, _rangeSize.Width];
            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    rangeBlock[i, j] = input[coords.Y + i, coords.X + j];
                }
            }
            /*   if (coords.Y > 0)
               {
                   Console.Write('x');
               }*/
            return rangeBlock;
        }


        static byte[,] transformToRangeBlock(byte[,] input, int X, int Y, int isometry)
        {
            byte[,] result = new byte[_rangeSize.Height, _rangeSize.Width];
            int[,] temp = new int[_rangeSize.Height, _rangeSize.Width];
            int width = result.GetLength(0);
            int height = result.GetLength(1);
            int weightX = _domainSize.Width / _rangeSize.Width;
            int weightY = _domainSize.Height / _rangeSize.Height;
            int numElem = weightX * weightY;
            for (int i = 0; i < _domainSize.Height; i++)
            {
                for (int j = 0; j < _domainSize.Width; j++)
                {
                    temp[i / weightY, j / weightX] += input[Y + i, X + j];
                }
            }


            /*
                    int newLum = (domainBlock[i * 2, j * 2] + domainBlock[i * 2 + 1, j * 2] + domainBlock[i * 2, j * 2+1] + domainBlock[i * 2 + 1, j * 2+1]) / 4;
                 
                    newLum=newLum+ shift;
                    if (newLum > 255)
                    {
                        temp[i, j] = 255;
                    }
                    else if (newLum < 0)
                    {
                        temp[i, j] = 0;

                    }
                    else
                    {
                        temp[i, j] = (byte)newLum;
                    }
                }
            } 
            */
            #region isometricTransformAndShift
            switch (isometry)
            {
                case 0:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[i, j] / numElem);

                            }
                        }

                    } break;
                case 1:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[height - j - 1, i] / numElem);

                            }
                        }
                    } break;
                case 2:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[height - i - 1, height - j - 1] / numElem);

                            }
                        }
                    } break;
                case 3:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[j, height - i - 1] / numElem);

                            }
                        }
                    } break;
                case 4:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[height - j - 1, height - i - 1] / numElem);

                            }
                        }
                    } break;
                case 5:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[i, height - j - 1] / numElem);

                            }
                        }
                    } break;
                case 6:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[j, i] / numElem);

                            }
                        }
                    } break;
                case 7:
                    {
                        for (int i = 0; i < height; i++)
                        {
                            for (int j = 0; j < width; j++)
                            {
                                result[i, j] = (byte)(temp[height - i - 1, j] / numElem);

                            }
                        }
                    } break;
            }
            #endregion

            return result;
        }





        static double MSEError(byte[,] rangeBlock, byte[,] transformedDomainBlock, double contrast, int shift)
        {
            double sum = 0;

            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    int val = (int)(transformedDomainBlock[i, j] * contrast + shift);
                    val=val < 255 ? (val < 0) ? (byte)0 : (byte)val : (byte)255;
                    sum += Math.Pow((rangeBlock[i, j] - val), 2);
                }
            }
            return sum / (_rangeSize.Height * _rangeSize.Height);
        }
        // static int getOptimalConAndBrigth(byte[,] rangeBlock, byte[,] transformedDomainBlock)
        

        public static void FullSearch(string filePath)
        {
            Bitmap img = new Bitmap(filePath);
            Bitmap result = new Bitmap(128, 128);
            int numRanges = img.Width * img.Height / (_rangeSize.Width * _rangeSize.Height);
            int numDomains = (img.Width - _domainSize.Width) * (img.Width - _domainSize.Width);
            byte[,] domainBlockBuffer = new byte[_domainSize.Height, _domainSize.Width];
            byte[,] transformedBlock = new byte[_rangeSize.Height, _rangeSize.Width];

            FractalMap[] Mappings = new FractalMap[numRanges];
            
            double currentError = double.MaxValue;
            GetPixelsFromBmp(img,0,0);

            for (int i = 0; i < numRanges; i++)
            {
                Console.WriteLine("{0} od {1}", i, numRanges);
                byte[,] rangeBlock = getRangeBlock(pixels, i);
                Mappings[i] = new FractalMap();
                currentError = double.MaxValue;
                int X = -1;
                int Y = 0;
                bool found = false;
                for (int j = 0; j < numDomains; j++)
                {
                    X++;

                    if (X > pixels.GetLength(1) - _domainSize.Width)
                    {
                        X = 0;
                        Y++;
                    }
                    if (Y > (pixels.GetLength(0) - _domainSize.Height)) break;

                    for (int k = 0; k < 8; k++)
                    {
                        transformedBlock = transformToRangeBlock(pixels, X, Y, k);
                        double contrast = 0;
                        int bright = 0;
                        getParameters(rangeBlock, transformedBlock, out contrast, out bright);
                        double error = MSEError(rangeBlock, transformedBlock, contrast, bright);
                        if (error < currentError)
                        {
                            currentError = error;
                            Mappings[i].X = X;
                            Mappings[i].Y = Y;
                            Mappings[i].Isometry = k;
                            Mappings[i].LumShift = bright;
                            Mappings[i].Contrast = contrast;
                        }
                    }
                    if (found) break;
                }
                Console.WriteLine(Mappings[i].ToString());
            }

            byte[,] outPixels = new byte[result.Height, result.Width];
            _domainSize = new Size(_domainSize.Width, _domainSize.Height);
            _rangeSize = new Size(_rangeSize.Width , _rangeSize.Height);
            for (int i = 0; i < result.Height; i++)
            {
                for (int j = 0; j < result.Width; j++)
                {

                    outPixels[i, j] = 128;

                }
            }
            int noIterations = 10;
            byte[,] temp = new byte[result.Height, result.Width];

            for (int iteration = 0; iteration < noIterations; iteration++)
            {



                for (int i = 0; i < Mappings.Length; i++)
                {

                    byte[,] rangeBuffer = transformToRangeBlock(outPixels, Mappings[i].X, Mappings[i].Y, Mappings[i].Isometry/* Mappings[i].LumShift*/);


                    for (int j = 0; j < _rangeSize.Height; j++)
                    {
                        for (int k = 0; k < _rangeSize.Width; k++)
                        {

                            int val = ((int)(rangeBuffer[j, k] * Mappings[i].Contrast) + Mappings[i].LumShift);
                            temp[j + ((i * _rangeSize.Width) / outPixels.GetLength(1)) * _rangeSize.Height
                                , k + ((i * _rangeSize.Width) % outPixels.GetLength(0))] = val < 255 ? (val < 0) ? (byte)0 : (byte)val : (byte)255;
                        }
                    }
                }
                byte[,] temp2 = outPixels;
                outPixels = temp;
                temp = temp2;

            }
            for (int i = 0; i < result.Height; i++)
            {
                for (int j = 0; j < result.Width; j++)
                {
                    result.SetPixel(j, i, Color.FromArgb(outPixels[i, j], outPixels[i, j], outPixels[i, j]));
                }
            }
            result.Save("IZLAZPOTPUNA.jpg", ImageFormat.Bmp);

        }

        static void getParameters12(byte[,] rangeBlock, byte[,] transformedDomainBlock, out double contrast, out int bright)
        {
            double avg1 = 0;
            double avg2 = 0;
            double alfa = 0;
            double beta = 0;
            int s = 0;
            int o = 0;
            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    avg1 += rangeBlock[i, j];
                    avg2 += 0.5 * transformedDomainBlock[i, j];
                }
            }
            avg1 /= (_rangeSize.Height * _rangeSize.Width);
            avg2 /= _rangeSize.Height * _rangeSize.Width;
            contrast = 0.5;
            bright = (int)(avg1 - avg2);


        }
        static void getParameters(byte[,] rangeBlock, byte[,] transformedDomainBlock, out double contrast, out int bright)
        {
            int sum1 = 0, sum2 = 0, sum12 = 0, sumSqueres2 = 0,sumSquares1=0, N;
            // double a,b;

            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    sum1 += rangeBlock[i, j];
                    sum2 += transformedDomainBlock[i, j];
                    sum12 += rangeBlock[i, j] * transformedDomainBlock[i, j];
                    sumSqueres2 += transformedDomainBlock[i, j] * transformedDomainBlock[i, j];
                    sumSquares1 += rangeBlock[i, j] * rangeBlock[i, j];

                }
            }
            N = _rangeSize.Width * _rangeSize.Width;
            if (N * sumSqueres2 - sum2 * sum2 == 0)
            {
                contrast = 1;
                bright = (sum1 - sum2) / N;
            }
            else
            {
                contrast = (double)(N * sum12 - sum1 * sum2) / (N * sumSqueres2 - sum2 * sum2);
             //   bright = (sumSqueres2 * sum1 - sum2 * sum12) / (N * sumSqueres2 - sum2 * sum2);
               // bright = (sum2* sum2 - sumSquares1) / (N * sumSqueres2 - sum2 * sum2); 
                //if (contrast < 0) contrast = 0;
                //int finContrast = (int)(0.5 + (contrast + 1) / 2.0 * (1 << 10));
                //if (finContrast < 0) finContrast = 0;
                //if (finContrast >= (1 << 10)) finContrast = 1023;
                //contrast = (double)finContrast / (1 << 10) * 2 - 1;
                if (contrast < 0) contrast = 0.0;
                int finContrast = (int)(0.5+contrast * (1 << 10));
                if (Math.Abs(finContrast) >= 1023)
                    finContrast = Math.Sign(finContrast) * 1023;
               //finContrast = -1024;
                contrast = (double)finContrast / (1 << 10);
                bright = (int)(sum1 - contrast * sum2) / N;
                if (Math.Abs(bright) > 255) bright = Math.Sign(bright) * 255;
                
            }

        }
        static int getParameters3(byte[,] rangeBlock, byte[,] transformedDomainBlock, out double contrast, out int bright)
        {
            double avg1 = 0;
            double avg2 = 0;
            double alfa = 0;
            double beta = 0;
            int s = 0;
            int o = 0;
            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    avg1 += rangeBlock[i, j];
                    avg2 += transformedDomainBlock[i, j];
                }
            }
            avg1 /= (_rangeSize.Height * _rangeSize.Width);
            avg2 /= _rangeSize.Height * _rangeSize.Width;
            for (int i = 0; i < _rangeSize.Height; i++)
            {
                for (int j = 0; j < _rangeSize.Width; j++)
                {
                    alfa += (rangeBlock[i, j] - avg1) * (transformedDomainBlock[i, j] - avg2);
                    beta += (transformedDomainBlock[i, j] - avg2) * (transformedDomainBlock[i, j] - avg2);
                }
            }
            if (beta == 0)
            {
                contrast = 1;
                bright =(int)(avg1-avg2);
                return 0;
            }

            contrast = alfa / beta;

            bright = (int)Math.Floor(avg1 - alfa / beta * avg2);



            //            return (int)((avg1-avg2)/(_rangeSize.Width*_rangeSize.Height));
            return 0;
        }
    }
}
