﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using FractalCompression;
namespace GUI
{
    public partial class Form1 : Form
    {

        frmView _picViewer;
        int currentStep;
        public Form1()
        {
            InitializeComponent();
            backgroundWorkerEncoder.WorkerReportsProgress = true;
            backgroundWorkerEncoder.ProgressChanged += new ProgressChangedEventHandler(backgroundWorkerEncoder_ProgressChanged);
            backgroundWorkerEncoder.RunWorkerCompleted += new RunWorkerCompletedEventHandler(backgroundWorkerEncoder_RunWorkerCompleted);

            txtInput.Text = @".\Ulaz\lenna_512.bmp";
            txtOutput.Text = @".\Izlaz\lenna_512.rg";
            txtMutationProbability.Text = "0.15";
            txtCrossoverProbability.Text = "0.8";
            txtGenNumber.Text = "5";
            txtPopulationSize.Text = "50";
            txtStagnation.Text = "7";
        }

        void backgroundWorkerEncoder_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            progressBarGenerations.Value = 0;
            btnStart.Enabled = true;
        }

        void backgroundWorkerEncoder_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            progressBarGenerations.Value = e.ProgressPercentage;
        }

        private void btnStart_Click(object sender, EventArgs e)
        {
             int popSize=0;
            int stagnationPeriod=0;
            double crossoverP=0;
            double mutationP=0;
            int generationNumber=0;
            try
            {
                popSize=Int32.Parse(txtPopulationSize.Text);
                stagnationPeriod=Int32.Parse(txtStagnation.Text);
                crossoverP=double.Parse(txtCrossoverProbability.Text);
                mutationP=double.Parse(txtMutationProbability.Text);
                if(crossoverP>1 || crossoverP<0 || mutationP>1 || mutationP<0)
                    throw new Exception();
                generationNumber=Int32.Parse(txtGenNumber.Text);
            }
            catch
            {
             MessageBox.Show("Greska u formatu parametara");
            }




            if (!File.Exists(txtInput.Text))
            {
                MessageBox.Show("ulazna datoteka ne postoji");
                return;
            }
            if (txtOutput.Text == "")
            {
                MessageBox.Show("potrebno odabrati izlaznu datoteku");
                return;
            }
            try
            {
                FractalEncoder encoder = new FractalEncoder(txtInput.Text, txtOutput.Text);
                encoder.CrossoverProbability = crossoverP;
                encoder.MutationProbability = mutationP;
                encoder.GenerationNumber = generationNumber;
                encoder.PopulationSize = popSize;
                encoder.StagnationThreshold = stagnationPeriod;
                encoder.ProgressChanged += new EventHandler(encoder_ProgressChanged);
                backgroundWorkerEncoder.RunWorkerAsync(encoder);
                btnStart.Enabled = false;
            }
            catch (Exception ex)
            {
                MessageBox.Show("Greska kod kodiranja");
                //backgroundWorkerEncoder.CancelAsync();
            }

          //  encoder.Encode();
            
            
        }

        void encoder_ProgressChanged(object sender, EventArgs e)
        {
            backgroundWorkerEncoder.ReportProgress( (sender as FractalEncoder).Progress);
        }

        private void backgroundWorkerEncoder_DoWork(object sender, DoWorkEventArgs e)
        {
            (e.Argument as FractalEncoder).Encode();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            ofdInput.ShowDialog();
            txtInput.Text = ofdInput.FileName;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            sfdOutput.ShowDialog();
            txtOutput.Text = sfdOutput.FileName;
        }

        private void btnBrowseRG_Click(object sender, EventArgs e)
        {
            ofdRG.ShowDialog();
            txtInputRG.Text = ofdRG.FileName;
            
        }

        private void btnBrowseBMP_Click(object sender, EventArgs e)
        {
            sfdBMP.ShowDialog();
            txtOutputBMP.Text = sfdBMP.FileName;
        }

        private void btnDecode_Click(object sender, EventArgs e)
        {
            if (_picViewer == null)
            {
                _picViewer = new frmView();
                _picViewer.FormClosed += new FormClosedEventHandler(_picViewer_FormClosed);
            }
            if (!File.Exists(txtInputRG.Text))
            {
                MessageBox.Show("Ulazna datoteka ne postoji");
                return;
            }
            if (txtOutputBMP.Text == "")
            {
                MessageBox.Show("Treba specificirati ime izlazne datoteke");
                return;
            }
            btnDecode.Enabled = false;
            btnStep.Enabled = false;
            try
            {
                FractalDecoder dec = new FractalDecoder(chkPostProccess.Checked);
                Bitmap result = dec.getBitmap(txtInputRG.Text, 10);

                result.Save(txtOutputBMP.Text);


                _picViewer.Show();

                _picViewer.Picture = result;
                btnDecode.Enabled = true;
                btnStep.Enabled = true;
            }
            catch (Exception ex)
            {

                MessageBox.Show("Greška kod dekodiranja");
            }
        }

        void _picViewer_FormClosed(object sender, FormClosedEventArgs e)
        {
            _picViewer = null;
        }

        private void btnView_Click(object sender, EventArgs e)
        {
           
            
        }

        private void button3_Click(object sender, EventArgs e)
        {
            if (_picViewer == null )
            {
                currentStep = 0;
                _picViewer = new frmView();
                _picViewer.FormClosed += new FormClosedEventHandler(_picViewer_FormClosed);
            }
            else
            {
                currentStep++;
            }
            if (!File.Exists(txtInputRG.Text))
            {
                MessageBox.Show("Ulazna datoteka ne postoji");
                return;
            }
            if (txtOutputBMP.Text == "")
            {
                MessageBox.Show("Treba specificirati ime izlazne datoteke");
                return;
            }
            btnDecode.Enabled = false;
            btnStep.Enabled = false;
            try
            {
                FractalDecoder dec = new FractalDecoder(chkPostProccess.Checked);
                Bitmap result = dec.getBitmap(txtInputRG.Text, currentStep);

                result.Save(txtOutputBMP.Text);


                _picViewer.Show();

                _picViewer.Picture = result;
                btnDecode.Enabled = true;
                btnStep.Enabled = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show("Greška prilikom dekodiranja");
            }


        }



        private void tabControl1_Selected(object sender, TabControlEventArgs e)
        {
            if (e.TabPage == tabDecode)
            {
                txtInputRG.Text = txtOutput.Text;
                string filename=txtInputRG.Text.Substring(txtInputRG.Text.LastIndexOf("\\")+1);
                txtOutputBMP.Text = ".\\Dekodiran\\"+filename.Replace(".rg", "") + ".bmp";
            }
            if (e.TabPage == tabCompare)
            {
                txtInCompare.Text = txtInput.Text;
                txtOutCompare.Text = txtOutputBMP.Text;
            }
        }

        private void btnInCompare_Click(object sender, EventArgs e)
        {
            ofdInCompare.ShowDialog();
            txtInCompare.Text = ofdInCompare.FileName;
        }

        private void btnOutCompare_Click(object sender, EventArgs e)
        {
            ofdOutCompare.ShowDialog();
            txtOutCompare.Text = ofdOutCompare.FileName;
        }

        private void btnCompare_Click(object sender, EventArgs e)
        {
            if (!File.Exists(txtInCompare.Text))
            {
                MessageBox.Show("Datoteka ne postoji");
                return;
            }
            if (!File.Exists(txtOutCompare.Text))
            {
                MessageBox.Show("Datoteka ne postoji");
                return;
            }
            try
            {
                txtPSNR.Text = Util.ComputePeakSignaltoNoise(txtInCompare.Text, txtOutCompare.Text).ToString();
                txtAvgError.Text = Util.ComputeAveragePixelError(txtInCompare.Text, txtOutCompare.Text).ToString();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Greška kod usporedbe");
            }


        }

  

       
    }
}
